# Copyright 2021 ONDEWO GmbH
#
# Licensed under the Apache License, Version 2.0 (the License);
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an AS IS BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
from contextlib import nullcontext as does_not_raise
from typing import Any, Dict, List

import pytest
from ondewo.nlu import intent_pb2, session_pb2

import ondewo_bpi.config as file_anchor  # noqa: F401
from ondewo_bpi.bpi_server import BpiServer
from ondewo_bpi.constants import DATE_FORMAT, QueryTriggers, SipTriggers
from ondewo_bpi.message_handler import MessageHandler, SingleMessageHandler

REPLACEMENT_STRING = "replacement string"


def create_text_response(text: str) -> Any:
    return session_pb2.DetectIntentResponse(
        response_id="d07e62f1-e652-473c-b445-c77a0ad5260d",
        query_result=session_pb2.QueryResult(
            fulfillment_messages=[intent_pb2.Intent.Message(text=intent_pb2.Intent.Message.Text(text=[text]))]
        ),
    )


@pytest.mark.parametrize(
    "original_message,processed_message",  # type: ignore
    [
        pytest.param("some random message", "some random message", marks=pytest.mark.xfail),
        (
            f"message with birthday: {datetime.datetime(1989, 4, 28).isoformat()}",
            f"message with birthday: {datetime.datetime(1989, 4, 28).strftime(DATE_FORMAT)}",
        ),
        (
            f"message with birthday: {datetime.datetime(1989, 4, 28, 22).isoformat()} in middle",
            f"message with birthday: {datetime.datetime(1989, 4, 28).strftime(DATE_FORMAT)} in middle",
        ),
        (
            f"message with multiple birthdays: {datetime.datetime(1989, 4, 28, 22).isoformat()} in {datetime.datetime(1989, 4, 20).isoformat()} middle",
            f"message with multiple birthdays: {datetime.datetime(1989, 4, 28).strftime(DATE_FORMAT)} in {datetime.datetime(1989, 4, 20).strftime(DATE_FORMAT)} middle",
        ),
    ],
)
def test_reformat_birthday_in_response(original_message, processed_message) -> None:
    response = create_text_response(original_message)
    response = MessageHandler.reformat_date(response)
    assert response.query_result.fulfillment_messages[0].text.text[0] == processed_message


@pytest.mark.parametrize(
    "original_message,expected_trigger",  # type: ignore
    [
        ("some random message", ([], [])),
        (
            f"message with trigger {SipTriggers.SIP_HANGUP.value}",
            ([SipTriggers.SIP_HANGUP.value], [SipTriggers.SIP_HANGUP.value],),
        ),
        (
            f"message with trigger {SipTriggers.SIP_HANGUP.value} in middle",
            ([SipTriggers.SIP_HANGUP.value], [SipTriggers.SIP_HANGUP.value],),
        ),
        (
            f"{SipTriggers.SIP_HANGUP.value} at start",
            ([SipTriggers.SIP_HANGUP.value], [SipTriggers.SIP_HANGUP.value],),
        ),
        (
            f"{SipTriggers.SIP_HUMAN_HANDOVER.value} at start",
            ([SipTriggers.SIP_HUMAN_HANDOVER.value], [SipTriggers.SIP_HUMAN_HANDOVER.value],),
        ),
        ("<SIP:SEND_NOW=('something')> at start", ([SipTriggers.SIP_SEND_NOW.value], ["something"]),),
        ("<SIP:PAUSE=('5s')> at start", ([SipTriggers.SIP_PAUSE.value], ["5s"])),
        (
            "<SIP:PAUSE=('5s')> at start <SIP:SEND_NOW=('something else')>",
            ([SipTriggers.SIP_SEND_NOW.value, SipTriggers.SIP_PAUSE.value], ["something else", "5s"],),
        ),
        (
            f"{SipTriggers.SIP_HUMAN_HANDOVER.value} triggers <SIP:SEND_NOW=('trigger1')> aplenty <SIP:SEND_NOW=('trigger2')> different kinds <SIP:PAUSE=('10s')>",
            (
                [SipTriggers.SIP_HUMAN_HANDOVER.value, SipTriggers.SIP_SEND_NOW.value, SipTriggers.SIP_PAUSE.value],
                [SipTriggers.SIP_HUMAN_HANDOVER.value, "trigger1", "trigger2", "10s"],
            ),
        ),
    ],
)
def test_check_response_for_triggers(original_message, expected_trigger) -> None:
    response = create_text_response(original_message)
    trigger = MessageHandler.check_for_triggers(response)
    assert trigger[0] == expected_trigger[0]
    assert trigger[1] == expected_trigger[1]


@pytest.mark.parametrize(
    "original_message,processed_message",  # type: ignore
    [
        ("some random message", "some random message"),
        (f"message with trigger {SipTriggers.SIP_HANGUP.value}", "message with trigger ",),
        (f"message with trigger {SipTriggers.SIP_HANGUP.value} in middle", "message with trigger  in middle",),
        (f"message with trigger {SipTriggers.SIP_HANGUP.value}", "message with trigger ",),
        (f"{SipTriggers.SIP_HANGUP.value} at start", " at start"),
        (f"{SipTriggers.SIP_HUMAN_HANDOVER.value} at start", " at start"),
        ("<SIP:SEND_NOW=('something')> at start", " at start"),
        ("<SIP:PAUSE=('5s')> at start", " at start"),
        ("<SIP:PAUSE=('5s')> at start <SIP:SEND_NOW=('something else')>", " at start ",),
        (
            f"{SipTriggers.SIP_HUMAN_HANDOVER.value} triggers <SIP:SEND_NOW=('trigger1')> aplenty <SIP:SEND_NOW=('trigger2')> different kinds <SIP:PAUSE=('10s')>",
            " triggers  aplenty  different kinds ",
        ),
    ],
)
def test_remove_triggers_from_reponse(original_message, processed_message) -> None:
    response = create_text_response(original_message)
    response = MessageHandler.remove_triggers_from_reponse(response)
    assert response.query_result.fulfillment_messages[0].text.text[0] == processed_message


@pytest.mark.parametrize(
    "original_message,search,replace,processed_message",  # type: ignore
    [
        ("some random message", "random", "waldo", "some waldo message"),
        ("message with hash #hash.hash to change", "#hash.hash", "", "message with hash  to change",),
    ],
)
def test_replace_text_in_response(original_message, search, replace, processed_message) -> None:
    response = create_text_response(original_message)
    response = MessageHandler.substitute_pattern(search, replace, response)
    assert response.query_result.fulfillment_messages[0].text.text[0] == processed_message


class TestMessageHandler(BpiServer):
    def __init__(self) -> None:
        super().__init__()
        self.register_trigger_handler(SipTriggers.SIP_HANGUP.value, self.trigger_function_1)
        self.register_trigger_handler(SipTriggers.SIP_HUMAN_HANDOVER.value, self.trigger_function_2)
        self.register_trigger_handler(SipTriggers.SIP_SEND_NOW.value, self.replacement_function_self)
        self.register_trigger_handler(QueryTriggers.REPLACEMENT_TRIGGER.value, self.replacement_function_string)

    def trigger_function_1(
        self,
        response: session_pb2.DetectIntentResponse,
        message: intent_pb2.Intent.Message,
        trigger: str,
        found_triggers: Dict[str, List[str]],
    ) -> intent_pb2.Intent.Message:
        raise KeyError

    def trigger_function_2(
        self,
        response: session_pb2.DetectIntentResponse,
        message: intent_pb2.Intent.Message,
        trigger: str,
        found_triggers: Dict[str, List[str]],
    ) -> intent_pb2.Intent.Message:
        raise ValueError

    def replacement_function_self(
        self,
        response: session_pb2.DetectIntentResponse,
        message: intent_pb2.Intent.Message,
        trigger: str,
        found_triggers: Dict[str, List[str]],
    ) -> intent_pb2.Intent.Message:
        for content in found_triggers[trigger]:
            SingleMessageHandler.substitute_pattern_in_message(message, trigger, content, once=True)
        return message

    def replacement_function_string(
        self,
        response: session_pb2.DetectIntentResponse,
        message: intent_pb2.Intent.Message,
        trigger: str,
        found_triggers: Dict[str, List[str]],
    ) -> intent_pb2.Intent.Message:
        SingleMessageHandler.substitute_pattern_in_message(message, trigger, REPLACEMENT_STRING)
        return message


@pytest.mark.parametrize(
    "original_message,expectation",  # type: ignore
    [
        ("some random message", does_not_raise()),
        (f"{SipTriggers.SIP_HANGUP.value}", pytest.raises(KeyError)),
        (f"{SipTriggers.SIP_HUMAN_HANDOVER.value}", pytest.raises(ValueError)),
        (f"{SipTriggers.SIP_SEND_NOW.value}", does_not_raise()),
        (f"{QueryTriggers.REPLACEMENT_TRIGGER.value}", does_not_raise()),
    ],
)
def test_trigger_functions(original_message, expectation):
    response = create_text_response(original_message)
    message_handler = TestMessageHandler()
    with expectation:
        response = message_handler.process_messages(response)


@pytest.mark.parametrize(
    "original_message,processed_message",  # type: ignore
    [
        ("some random message", "some random message"),
        ("<SIP:SEND_NOW=('something')> at start", "something at start"),
        ("<SIP:PAUSE=('5s')> at start", " at start"),
        (
            "triggers <SIP:SEND_NOW=('trigger1')> aplenty <SIP:SEND_NOW=('trigger2')> different kinds <SIP:PAUSE=('10s')>",
            "triggers trigger2 aplenty trigger1 different kinds ",
        ),
        ("<c-examination.appointment_date>", REPLACEMENT_STRING),
        (
            "<c-examination.appointment_date> <c-examination.appointment_date>",
            f"{REPLACEMENT_STRING} {REPLACEMENT_STRING}",
        ),
    ],
)
def test_trigger_functions_replacement(original_message, processed_message):
    response = create_text_response(original_message)
    message_handler = TestMessageHandler()
    response = message_handler.process_messages(response)
    assert response.query_result.fulfillment_messages[0].text.text[0] == processed_message



# Copyright 2021 ONDEWO GmbH
#
# Licensed under the Apache License, Version 2.0 (the License);
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an AS IS BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import time
from typing import List, Tuple, Coroutine

import grpc
from ondewo.nlu import session_pb2
from ondewo.nlu.session_pb2 import DetectIntentResponse, DetectIntentRequest
from ondewo.qa import qa_pb2, qa_pb2_grpc
from ondewo.logging.decorators import Timer
from ondewo.logging.logger import logger_console

from ondewo_bpi_qa.bpi_qa_base_server import BpiQABaseServer
from ondewo_bpi_qa.config import (
    QA_LANG,
    QA_ACTIVE,
    QA_MAX_ANSWERS,
    QA_THRESHOLD_READER,
    QA_THRESHOLD_RETRIEVER,
    QA_HOST,
    QA_PORT,
    SESSION_TIMEOUT_MINUTES,
)


class QAServer(BpiQABaseServer):
    def __init__(self) -> None:
        super().__init__()
        self.qa_client_stub = qa_pb2_grpc.QAStub(channel=grpc.insecure_channel(f"{QA_HOST}:{QA_PORT}"))
        self.loops = {}  # type: ignore

    def serve(self) -> None:
        super().serve()

    @Timer(log_arguments=False)
    def DetectIntent(self, request: DetectIntentRequest, context: grpc.ServicerContext) -> DetectIntentResponse:
        self.check_session_id(request)
        response, response_name = self.handle_async(request)

        if response_name == "cai_response":
            # Process CAI response
            response = self.process_messages(response)
            response = self.process_intent_handler(response)
        return response

    def check_session_id(self, request: DetectIntentRequest) -> None:
        session_pop_timeout: int = SESSION_TIMEOUT_MINUTES * 60

        for session in self.loops.copy():
            if time.time() - self.loops[session]["timestamp"] > session_pop_timeout:  # type: ignore
                logger_console.warning(f"Popping old session: {session}.")
                loop = self.loops.pop(session)
                loop["loop"].stop()
                loop["loop"].close()

        if request.session not in self.loops.keys():
            self.loops[request.session] = {
                "loop": asyncio.new_event_loop(),
                "timestamp": time.time(),
            }
            logger_console.warning(
                f"New session in bpi: {request.session}. {len(self.loops)} sessions currently stored."
            )

    @Timer(log_arguments=False)
    def handle_async(self, request: DetectIntentRequest,) -> Tuple[DetectIntentResponse, str]:
        tasks: List[Coroutine] = [self.send_to_cai(request)]
        if QA_ACTIVE:
            tasks.append(self.send_to_qa(request))

        while len(tasks):
            logger_console.warning(f"Starting async loop with {len(tasks)} tasks of type: {type(tasks)}")
            try:
                finished, tasks = self.loops[request.session]["loop"].run_until_complete(
                    asyncio.wait(tasks, return_when=asyncio.FIRST_COMPLETED)  # type: ignore
                )
            except Exception:
                logger_console.exception("Task returned an exception!")
                return DetectIntentResponse(), "exception"

            # Allow the cai_response to return early if it finishes first
            for task in finished:
                result = task.result()
                if result[1] == "cai_response":
                    cai_response = result[0]
                    intent_name_cai = cai_response.query_result.intent.display_name
                    if intent_name_cai != "Default Fallback Intent" or not QA_ACTIVE:
                        logger_console.warning("CAI response good, returning early")
                        return cai_response, "cai_response"
                # If the QA response finishes first, save it for later
                else:
                    assert result[1] == "qa_response", "Somehow a different response snuck in!"
                    qa_response = result[0]

        qa_confidence = qa_response.query_result.query_result.intent_detection_confidence
        logger_console.warning(f"QA confidence is {qa_confidence}, cutoff is {QA_THRESHOLD_READER}")
        messages = qa_response.query_result.query_result.fulfillment_messages
        if messages:
            return qa_response.query_result, "qa_response"
        logger_console.warning("No response from QA, passing back Default Fallback.")
        return cai_response, "cai_response"

    async def send_to_qa(self, request: DetectIntentRequest,) -> Tuple[DetectIntentResponse, str]:
        text = request.query_input.text.text
        qa_request = qa_pb2.GetAnswerRequest(
            session_id=request.session,
            text=session_pb2.TextInput(text=text, language_code=f"{QA_LANG}"),
            max_num_answers=QA_MAX_ANSWERS,
            threshold_reader=QA_THRESHOLD_READER,
            threshold_retriever=QA_THRESHOLD_RETRIEVER,
        )

        logger_console.warning(
            {
                "message": f"QA-GetAnswerRequest to QA, text input: {text}",
                "content": text,
                "text": text,
                "tags": ["text"],
            }
        )
        qa_response: DetectIntentResponse = await self.loops[request.session]["loop"].run_in_executor(
            None, self.qa_client_stub.GetAnswer, qa_request,
        )
        # intent_name_qa = qa_response.query_result.intent.display_name
        logger_console.warning({"message": "QA-DetectIntentResponse from QA", "tags": ["text"]})
        return qa_response, "qa_response"

    async def send_to_cai(self, request: DetectIntentRequest,) -> Tuple[DetectIntentResponse, str]:
        text = request.query_input.text.text
        logger_console.warning(
            {
                "message": f"CAI-DetectIntentRequest to CAI, text input: {text}",
                "content": text,
                "text": text,
                "tags": ["text"],
            }
        )
        cai_response: DetectIntentResponse = await self.loops[request.session]["loop"].run_in_executor(
            None, self.client.services.sessions.detect_intent, request,
        )
        intent_name_cai = cai_response.query_result.intent.display_name
        logger_console.warning(
            {
                "message": f"CAI-DetectIntentResponse from CAI, intent_name_cai: {intent_name_cai}",
                "content": intent_name_cai,
                "intent_name_cai": intent_name_cai,
                "tags": ["text"],
            }
        )
        return cai_response, "cai_response"



# Copyright 2021 ONDEWO GmbH
#
# Licensed under the Apache License, Version 2.0 (the License);
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an AS IS BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional
import os

TIMEOUT_MINUTES: int = int(os.getenv("TIMEOUT_MINUTES", "20"))
PROVIDER_STT: Optional[str] = os.getenv("SIP_PROVIDER_STT")
PROVIDER_TTS: Optional[str] = os.getenv("SIP_PROVIDER_TTS")
LANG: Optional[str] = os.getenv("SIP_LANG")



# Copyright 2021 ONDEWO GmbH
#
# Licensed under the Apache License, Version 2.0 (the License);
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an AS IS BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from typing import Tuple, Optional

from ondewo.qa.client import Client
from ondewo.qa.client_config import ClientConfig
from ondewo.logging.logger import logger_console, logger

QA_HOST: str = os.getenv("QA_HOST", "172.17.0.1")
QA_PORT: str = os.getenv("QA_PORT", "50052")
QA_LANG: str = os.getenv("QA_LANG", "de")

SESSION_TIMEOUT_MINUTES: int = 20

QA_MAX_ANSWERS: int = int(os.getenv("QA_MAX_ANSWERS", "3"))
QA_THRESHOLD_READER: float = float(os.getenv("QA_THRESHOLD_READER", "0.5"))
QA_THRESHOLD_RETRIEVER: float = float(os.getenv("QA_THRESHOLD_RETRIEVER", "0.5"))
QA_ACTIVE: bool = True if os.getenv("QA_ACTIVE", "False") == "True" else False
QA_SECURE: Optional[str] = os.getenv("QA_SECURE", "False")

client_configuration_str = (
    "\nqa-client configuration:\n"
    + f"   Secure: {QA_SECURE}\n"
    + f"   Host: {QA_HOST}\n"
    + f"   Port: {QA_PORT}\n"
    + f"   Language: {QA_LANG}\n"
    + f"   Number of answers per query: {QA_MAX_ANSWERS}\n"
    + f"   Reader threshold: {QA_THRESHOLD_READER}\n"
    + f"   Retriever threshold: {QA_THRESHOLD_RETRIEVER}\n"
    + f"   Is active?: {QA_ACTIVE}\n"
)
logger_console.info(client_configuration_str)


class QAClientProvider:
    """
    provide a central qa-client instance to the bpi server without building it on import
    """

    def __init__(self) -> None:
        self.config = None
        self.client = None
        self._built = False

    def instantiate_client(self, qa_port: str = "") -> Tuple[ClientConfig, Client]:
        if qa_port == "":
            qa_port = QA_PORT

        if QA_SECURE:
            logger.warning("Secure connection not possible for Question&Answering.")
            logger.warning("Using insecure connection instead...")

        logger.info("configuring INSECURE connection")
        config = ClientConfig(host=QA_HOST, port=qa_port,)
        client = Client(config=config, use_secure_channel=False)
        return config, client

    def get_client(self, qa_port: str = "") -> Client:
        if not self._built:
            self.config, self.client = self.instantiate_client(qa_port=qa_port)
            self._built = True
        return self.client



# Copyright 2021 ONDEWO GmbH
#
# Licensed under the Apache License, Version 2.0 (the License);
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an AS IS BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABCMeta, abstractmethod

from ondewo.qa.client import Client as QAClient

from ondewo_bpi.autocoded.qa_grpc_autocode import AutoQAServicer


class BpiQAServices(AutoQAServicer):
    __metaclass__ = ABCMeta

    @property
    @abstractmethod
    def qa_client(self) -> QAClient:
        pass



# Copyright 2021 ONDEWO GmbH
#
# Licensed under the Apache License, Version 2.0 (the License);
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an AS IS BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time
from typing import Dict, Optional, Tuple, Any

import grpc
from ondewo.nlu import intent_pb2, session_pb2
from ondewo.logging.decorators import Timer
from ondewo.logging.logger import logger_console
from ondewo.sip.client_config import SipClientConfig
from ondewo.sip.client import SipClient

import ondewo_bpi.helpers as helpers
from ondewo_bpi.bpi_server import BpiServer
from ondewo_bpi_sip.config import TIMEOUT_MINUTES


def get_flask_url_from_session_id(session: str) -> str:
    identifying_info = session.split("XXX")[1].split("-")
    return f"http://{identifying_info[0]}:{identifying_info[1]}/{identifying_info[2]}"


def get_sip_host_name_port_from_session_id(session: str) -> Tuple[str, Optional[str], str]:
    identifying_info = session.split("XXX")[1].split("-")
    if "None" in identifying_info[1]:
        return identifying_info[0], None, identifying_info[2]
    return identifying_info[0], identifying_info[1], identifying_info[2]


class SipServer(BpiServer):
    def __init__(self) -> None:
        super().__init__()
        self.session_information: Dict[str, Dict[str, Any]] = {
            "session_uuid": {"client": None, "timestamp": 0.0}
        }

    def DetectIntent(
        self, request: session_pb2.DetectIntentRequest, context: grpc.ServicerContext
    ) -> session_pb2.DetectIntentResponse:
        self.check_session(request)
        return super().DetectIntent(request=request, context=context)  # type: ignore

    @Timer(log_arguments=False)
    def check_session(self, request: session_pb2.DetectIntentRequest) -> None:
        if request.session not in self.session_information.keys():
            host, port, name = get_sip_host_name_port_from_session_id(request.session)
            if None in [host, port, name]:
                logger_console.warning(
                    "WARNING: Sip client not initialized correctly!"
                    " No responses will be sent!"
                )
                self.session_information[request.session] = {
                    "client": None,
                    "timestamp": time.time()
                }
            else:
                logger_console.warning(
                    f"Sip host: {host}:{port}/{name}"
                )
                config: SipClientConfig = SipClientConfig(
                    host=host,
                    port=port,
                    name=name,
                )

                sip_client: SipClient = SipClient(config=config)
                self.session_information[request.session] = {
                    "client": sip_client,
                    "timestamp": time.time()
                }
            logger_console.warning(
                "New session in bpi."
                f" {len(self.session_information)} sessions currently stored."
            )

        for session in self.session_information.copy():
            current_age = time.time() - self.session_information[session]["timestamp"]
            if current_age > (TIMEOUT_MINUTES * 60):
                logger_console.warning(
                    f"Popping old session: session = {session}"
                    f" timestamp = {self.session_information[session]['timestamp']}"
                    f" age = {current_age},"
                    f" timeout is {TIMEOUT_MINUTES} minutes."
                )
                self.session_information.pop(session)

    def quicksend_to_api(
        self, response: session_pb2.DetectIntentResponse, message: Optional[intent_pb2.Intent.Message], count: int,
    ) -> None:
        client = self.session_information[helpers.get_session_from_response(response)]["client"]
        if not client:
            logger_console.warning(
                "WARNING: Sip client not initialized correctly! No responses will be sent!"
            )
            return
        if not message:
            client.services.text_to_speech.send_text_get_filename(  # type: ignore
                text_to_send="", session_id=helpers.get_session_from_response(response), response_type="final"
            )
            return
        if not message.HasField("text"):
            client.services.text_to_speech.send_text_get_filename(  # type: ignore
                text_to_send="", session_id=helpers.get_session_from_response(response), response_type="final"
            )
            return
        if message.HasField("card"):
            client.services.text_to_speech.send_text_get_filename(  # type: ignore
                text_to_send="", session_id=helpers.get_session_from_response(response), response_type="final"
            )
            return
        is_final: bool = ((count + 1) == len(response.query_result.fulfillment_messages))
        client.services.text_to_speech.send_text_get_filename(  # type: ignore
            text_to_send=message.text.text[0],
            session_id=helpers.get_session_from_response(response),
            response_type="final" if is_final else "partial",
        )



# Copyright 2021 ONDEWO GmbH
#
# Licensed under the Apache License, Version 2.0 (the License);
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an AS IS BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum

DATE_FORMAT: str = r"%d.%m.%4Y"
TIME_FORMAT: str = r"%H:%M:%S"


class SipTriggers(Enum):
    SIP_HANGUP: str = "<SIP:HANGUP>"
    SIP_HUMAN_HANDOVER: str = "<SIP:HUMAN_HANDOVER>"
    SIP_SEND_NOW: str = "<SIP:SEND_NOW=(.*?)>"
    SIP_PAUSE: str = "<SIP:PAUSE=(.*?)>"


class QueryTriggers(Enum):
    REPLACEMENT_TRIGGER: str = r"<(\w*-\w*\.\w*)>"



# Copyright 2021 ONDEWO GmbH
#
# Licensed under the Apache License, Version 2.0 (the License);
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an AS IS BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ondewo.nlu.client import Client as NLUClient
from ondewo.qa import qa_pb2, qa_pb2_grpc
from ondewo.qa.client import Client as QAClient

from ondewo_bpi.bpi_server import BpiServer
from ondewo_bpi_qa.bpi_qa_services import BpiQAServices
from ondewo_bpi_qa.config import QAClientProvider


class BpiQABaseServer(
    BpiServer,
    BpiQAServices,
):
    @property
    def client(self) -> NLUClient:
        return self._client

    @client.setter
    def client(self, value: NLUClient) -> None:
        self._client = value

    @property
    def qa_client(self) -> QAClient:
        return self._qa_client

    @qa_client.setter
    def qa_client(self, value: QAClient) -> None:
        self._qa_client = value

    def __init__(self) -> None:
        super(BpiQABaseServer, self).__init__()
        self.qa_client = QAClientProvider().get_client()
        self.services_descriptors.append(qa_pb2.DESCRIPTOR.services_by_name["QA"].full_name)  # type: ignore

    def _add_services(self) -> None:
        super(BpiQABaseServer, self)._add_services()
        qa_pb2_grpc.add_QAServicer_to_server(self, self.server)



# Copyright 2021 ONDEWO GmbH
#
# Licensed under the Apache License, Version 2.0 (the License);
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an AS IS BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time
from concurrent import futures
from typing import List

import grpc
from grpc_reflection.v1alpha import reflection
from ondewo.nlu import (
    agent_pb2,
    entity_type_pb2,
    aiservices_pb2,
    project_role_pb2,
    intent_pb2,
    session_pb2,
    session_pb2_grpc,
    user_pb2,
    user_pb2_grpc,
    context_pb2,
    context_pb2_grpc,
    agent_pb2_grpc,
    entity_type_pb2_grpc,
    intent_pb2_grpc,
    aiservices_pb2_grpc,
    project_role_pb2_grpc,
)
from ondewo.nlu.client import Client as NLUClient
from ondewo.logging.decorators import Timer
from ondewo.logging.logger import logger, logger_console

from ondewo_bpi.bpi_services import BpiSessionsServices, BpiUsersServices, BpiContextServices, \
    BpiAgentsServices, BpiEntityTypeServices, BpiAiServicesServices, BpiIntentsServices, \
    BpiProjectRolesServices
from ondewo_bpi.config import PORT, CentralClientProvider


class BpiServer(
    BpiAgentsServices,
    BpiAiServicesServices,
    BpiContextServices,
    BpiEntityTypeServices,
    BpiIntentsServices,
    BpiProjectRolesServices,
    BpiSessionsServices,
    BpiUsersServices,
):
    @property
    def client(self) -> NLUClient:
        return self._client

    @client.setter
    def client(self, value: NLUClient) -> None:
        self._client = value

    def __init__(self) -> None:
        super().__init__()
        self.client = CentralClientProvider().get_client()
        self.server = None
        self.services_descriptors: List[str] = [
            agent_pb2.DESCRIPTOR.services_by_name["Agents"].full_name,  # type: ignore
            context_pb2.DESCRIPTOR.services_by_name["Contexts"].full_name,  # type: ignore
            entity_type_pb2.DESCRIPTOR.services_by_name["EntityTypes"].full_name,  # type: ignore
            intent_pb2.DESCRIPTOR.services_by_name["Intents"].full_name,  # type: ignore
            session_pb2.DESCRIPTOR.services_by_name["Sessions"].full_name,  # type: ignore
            aiservices_pb2.DESCRIPTOR.services_by_name["AiServices"].full_name,  # type: ignore
            project_role_pb2.DESCRIPTOR.services_by_name["ProjectRoles"].full_name,  # type: ignore
            user_pb2.DESCRIPTOR.services_by_name["Users"].full_name,  # type: ignore
        ]

    def _setup_reflection(self) -> None:
        reflection.enable_server_reflection(service_names=self.services_descriptors, server=self.server)

    def _add_services(self) -> None:
        agent_pb2_grpc.add_AgentsServicer_to_server(self, self.server)
        context_pb2_grpc.add_ContextsServicer_to_server(self, self.server)
        entity_type_pb2_grpc.add_EntityTypesServicer_to_server(self, self.server)
        intent_pb2_grpc.add_IntentsServicer_to_server(self, self.server)
        session_pb2_grpc.add_SessionsServicer_to_server(self, self.server)
        aiservices_pb2_grpc.add_AiServicesServicer_to_server(self, self.server)
        project_role_pb2_grpc.add_ProjectRolesServicer_to_server(self, self.server)
        user_pb2_grpc.add_UsersServicer_to_server(self, self.server)

    @Timer(log_arguments=False)
    def _setup_server(self) -> None:
        logger.info("attempting to setup server...")
        self.server = grpc.server(futures.ThreadPoolExecutor(max_workers=10))
        self._add_services()
        self._setup_reflection()
        self.server.add_insecure_port(f"[::]:{PORT}")  # type: ignore
        logger.info(f"SERVING SERVER AT SERVING PORT {PORT}")
        self.server.start()  # type: ignore

    def serve(self) -> None:
        logger_console.info(f"attempting to start server on port {PORT}")
        self._setup_server()
        logger_console.warning({"message": f"Server started on port {PORT}", "content": PORT})
        logger_console.warning(
            {
                "message": f"using intent handlers dict {list(self.intent_handlers.keys())}",
                "content": list(self.intent_handlers.keys()),
            }
        )
        try:
            while True:
                time.sleep(10)
        except KeyboardInterrupt:
            logger_console.info("Keyboard interrupt, shutting down")
        logger_console.info({"message": "server shut down", "tags": ["timing"]})



# Copyright 2021 ONDEWO GmbH
#
# Licensed under the Apache License, Version 2.0 (the License);
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an AS IS BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABCMeta, abstractmethod
from typing import Dict, Callable, List, Optional

import grpc
from ondewo.nlu import session_pb2, intent_pb2, user_pb2, context_pb2
from ondewo.nlu.client import Client as NLUClient
from ondewo.logging.decorators import Timer
from ondewo.logging.logger import logger_console

from ondewo_bpi.autocoded.agent_grpc_autocode import AutoAgentsServicer
from ondewo_bpi.autocoded.aiservices_grpc_autocode import AutoAiServicesServicer
from ondewo_bpi.autocoded.context_grpc_autocode import AutoContextsServicer
from ondewo_bpi.autocoded.entity_type_grpc_autocode import AutoEntityTypesServicer
from ondewo_bpi.autocoded.intent_grpc_autocode import AutoIntentsServicer
from ondewo_bpi.autocoded.project_role_grpc_autocode import AutoProjectRolesServicer
from ondewo_bpi.autocoded.session_grpc_autocode import AutoSessionsServicer
from ondewo_bpi.autocoded.user_grpc_autocode import AutoUsersServicer
from ondewo_bpi.constants import SipTriggers, QueryTriggers
from ondewo_bpi.message_handler import MessageHandler, SingleMessageHandler

from ondewo_bpi.helpers import get_session_from_response


class BpiSessionsServices(AutoSessionsServicer):
    __metaclass__ = ABCMeta

    @property
    @abstractmethod
    def client(self) -> NLUClient:
        pass

    def __init__(self) -> None:
        self.intent_handlers: Dict[str, Callable] = {}
        self.trigger_handlers: Dict[str, Callable] = {
            i.value: self.trigger_function_not_implemented for i in [*SipTriggers, *QueryTriggers]
        }

    def register_intent_handler(self, intent_name: str, handler: Callable) -> None:
        self.intent_handlers[intent_name] = handler

    def register_trigger_handler(self, trigger: str, handler: Callable) -> None:
        self.trigger_handlers[trigger] = handler

    def trigger_function_not_implemented(
            self,
            response: session_pb2.DetectIntentResponse,
            message: intent_pb2.Intent.Message,
            trigger: str,
            found_triggers: Dict[str, List[str]],
    ) -> None:
        logger_console.warning(
            {
                "message": f"no function for the trigger {trigger}, please subclass and implement",
                "trigger": trigger,
                "content": found_triggers[trigger],
            }
        )

    def DetectIntent(
            self, request: session_pb2.DetectIntentRequest, context: grpc.ServicerContext
    ) -> session_pb2.DetectIntentResponse:
        try:
            text = request.query_input.text.text
        except Exception:
            logger_console.exception("something wrong in the bpi")
            text = "error"
        logger_console.warning(
            {
                "message": f"CAI-DetectIntentRequest to CAI, text input: {text}",
                "content": text,
                "text": text,
                "tags": ["text"],
            }
        )
        cai_response = self.perform_detect_intent(request)
        intent_name = cai_response.query_result.intent.display_name
        logger_console.warning(
            {
                "message": f"CAI-DetectIntentResponse from CAI, intent_name: {intent_name}",
                "content": intent_name,
                "intent_name": intent_name,
                "session_id": get_session_from_response(cai_response),
                "tags": ["text"],
            }
        )

        cai_response = self.process_messages(cai_response)
        return self.process_intent_handler(cai_response)

    @Timer(log_arguments=False, recursive=True)
    def perform_detect_intent(self,
                              request: session_pb2.DetectIntentRequest, ) -> session_pb2.DetectIntentResponse:
        return self.client.services.sessions.detect_intent(request)

    @Timer(log_arguments=False, recursive=True)
    def process_messages(self,
                         response: session_pb2.DetectIntentResponse, ) -> session_pb2.DetectIntentResponse:
        new_response = None
        for j, message in enumerate(response.query_result.fulfillment_messages):
            found_triggers = MessageHandler.get_triggers(message, get_session_from_response(response))

            for found_trigger in found_triggers:
                new_response = self.trigger_handlers[found_trigger](response, message, found_trigger,
                                                                    found_triggers)
                if new_response:
                    if not new_response.response_id == response.response_id:
                        return new_response

            for found_trigger in found_triggers:
                SingleMessageHandler.substitute_pattern_in_message(message, found_trigger, "")

            self.quicksend_to_api(response, message, j)
        if not len(response.query_result.fulfillment_messages):
            self.quicksend_to_api(response, None, 0)
        return response

    def quicksend_to_api(
            self, response: session_pb2.DetectIntentResponse, message: Optional[intent_pb2.Intent.Message],
            count: int
    ) -> None:
        logger_console.warning({"message": "quicksend_to_api not written, please subclass and implement"})

    @Timer(log_arguments=False, recursive=True)
    def process_intent_handler(
            self, cai_response: session_pb2.DetectIntentResponse
    ) -> session_pb2.DetectIntentResponse:
        intent_name = cai_response.query_result.intent.display_name
        handler: Optional[Callable] = self.intent_handlers.get(intent_name)
        if handler is not None:
            cai_response = handler(cai_response)
            text = [i.text.text for i in cai_response.query_result.fulfillment_messages]
            logger_console.warning(
                {
                    "message": f"BPI-DetectIntentResponse from BPI with text: {text}",
                    "content": text,
                    "text": text,
                    "tags": ["text", "clean"],
                }
            )
        return cai_response


class BpiUsersServices(AutoUsersServicer):
    __metaclass__ = ABCMeta

    @property
    @abstractmethod
    def client(self) -> NLUClient:
        pass

    def Login(self, request: user_pb2.LoginRequest, context: grpc.ServicerContext) -> user_pb2.LoginResponse:
        logger_console.info("login request handled by bpi")
        return user_pb2.LoginResponse(auth_token=self.client.services.users.metadata[0][1])


class BpiContextServices(AutoContextsServicer):
    __metaclass__ = ABCMeta

    @property
    @abstractmethod
    def client(self) -> NLUClient:
        pass

    def CreateContext(
            self, request: context_pb2.CreateContextRequest, context: grpc.ServicerContext
    ) -> context_pb2.Context:
        logger_console.info("passing create context request on to CAI")
        return self.client.services.contexts.create_context(request=request)


class BpiAgentsServices(AutoAgentsServicer):
    __metaclass__ = ABCMeta

    @property
    @abstractmethod
    def client(self) -> NLUClient:
        pass


class BpiEntityTypeServices(AutoEntityTypesServicer):
    __metaclass__ = ABCMeta

    @property
    @abstractmethod
    def client(self) -> NLUClient:
        pass


class BpiAiServicesServices(AutoAiServicesServicer):
    __metaclass__ = ABCMeta

    @property
    @abstractmethod
    def client(self) -> NLUClient:
        pass


class BpiIntentsServices(AutoIntentsServicer):
    __metaclass__ = ABCMeta

    @property
    @abstractmethod
    def client(self) -> NLUClient:
        pass


class BpiProjectRolesServices(AutoProjectRolesServicer):
    __metaclass__ = ABCMeta

    @property
    @abstractmethod
    def client(self) -> NLUClient:
        pass



# Copyright 2021 ONDEWO GmbH
#
# Licensed under the Apache License, Version 2.0 (the License);
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an AS IS BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
import re
from typing import Any, Dict, List, Optional, Tuple, Union

from google.protobuf.json_format import MessageToJson
from ondewo.nlu import context_pb2, intent_pb2, session_pb2
from ondewo.logging.logger import logger_console

from ondewo_bpi.constants import DATE_FORMAT, QueryTriggers, SipTriggers


def create_parameter_dict(my_dict: Dict) -> Optional[Dict[str, context_pb2.Context.Parameter]]:
    assert isinstance(my_dict, dict) or my_dict is None, "parameter must be a dict or None"
    if my_dict is not None:
        return {
            key: context_pb2.Context.Parameter(
                display_name=key,
                value=my_dict[key]
            )
            for key in my_dict
        }
    return None


class MessageHandler:
    """
    This class takes a grpc response and decomposes the message into multiple parts, processes them asynchronously and
    sends the reponses to all available enpoints. Once the processing is complete, the message is reassembled and sent
    back.

    The subclasser should implment 'quicksend_to_api' if they want messages to be send to an endpoint out of turn (aka
    before the whole response is finished), and should implement functions for each of the triggers that they want to
    handle and update self.trigger_function_map with the new functions.
    """

    @staticmethod
    def get_triggers(message: intent_pb2.Intent.Message, session_id: Optional[str] = None) -> Dict[str, List[str]]:
        found_triggers: Dict[str, List[str]] = {}
        for trigger in SipTriggers:
            if SingleMessageHandler.check_message_for_pattern(message, trigger.value):
                content = SingleMessageHandler.get_pattern_from_message(message, trigger.value)
                found_triggers[trigger.value] = content
        for qtrigger in QueryTriggers:
            if SingleMessageHandler.check_message_for_pattern(message, qtrigger.value):
                content = SingleMessageHandler.get_pattern_from_message(message, qtrigger.value)
                found_triggers[qtrigger.value] = content
        if len(found_triggers):
            logger_console.info({"message": f"Found triggers: {found_triggers}", "found_triggers": found_triggers, "session_id": session_id})
        return found_triggers

    @staticmethod
    def check_for_pattern(pattern: str, response: session_pb2.DetectIntentResponse,) -> bool:
        for message in response.query_result.fulfillment_messages:
            if not len(message.text.text):
                continue
            if SingleMessageHandler.check_message_for_pattern(message, pattern):
                return True
        return False

    @staticmethod
    def substitute_pattern(
        pattern: str, replace: str, response: session_pb2.DetectIntentResponse,
    ) -> session_pb2.DetectIntentResponse:
        logger_console.info({"message": "replacing text in response", "pattern": pattern, "replace": replace})
        for j, message in enumerate(response.query_result.fulfillment_messages):
            SingleMessageHandler.substitute_pattern_in_message(message, pattern, replace)
        return response

    @staticmethod
    def reformat_date(response: session_pb2.DetectIntentResponse,) -> session_pb2.DetectIntentResponse:
        logger_console.info("reformatting date in response")
        for message in response.query_result.fulfillment_messages:
            if not len(message.text.text):
                continue
            SingleMessageHandler.reformat_date_in_message(message)
        return response

    @staticmethod
    def check_for_triggers(response: session_pb2.DetectIntentResponse,) -> Tuple[List[str], List[str]]:
        found_triggers: List[str] = []
        trigger_content: List[str] = []
        for j, message in enumerate(response.query_result.fulfillment_messages):
            if not len(message.text.text):
                continue
            for trigger in SipTriggers:
                if SingleMessageHandler.check_message_for_pattern(message, trigger.value):
                    found_triggers.append(trigger.value)
                    trigger_content.extend(SingleMessageHandler.get_pattern_from_message(message, trigger.value))
        return found_triggers, trigger_content  # type: ignore

    @staticmethod
    def remove_triggers_from_reponse(response: session_pb2.DetectIntentResponse,) -> session_pb2.DetectIntentResponse:
        for j, message in enumerate(response.query_result.fulfillment_messages):
            if not len(message.text.text):
                continue
            for trigger in SipTriggers:
                SingleMessageHandler.substitute_pattern_in_message(message, trigger.value, "")
        return response


class ParameterMethods:
    @staticmethod
    def get_context(response: session_pb2.DetectIntentResponse, context_name: str) -> Optional[context_pb2.Context]:
        logger_console.info({"message": "searching for context", "content": context_name, "tags": ["contexts"]})
        context = [c for c in response.query_result.output_contexts if c.name == context_name]
        logger_console.info(
            {"message": "found context", "content": MessageToJson(context[0]) if len(context) else "None"}
        )
        return context[0] if len(context) else None

    @staticmethod
    def get_param(response: session_pb2.DetectIntentResponse, param_name: str, context_name: str,) -> Any:
        logger_console.info({"message": "getting param", "content": context_name, "tags": ["parameters"]})
        context = ParameterMethods.get_context(response=response, context_name=context_name)
        if context is None:
            return None

        params = list(context.parameters)
        if param_name in params:
            returned_param = context.parameters[param_name]
        else:
            returned_param = "None"

        try:
            returned_param = MessageToJson(returned_param)
        except AttributeError:
            pass

        logger_console.info({"message": "found param", "content": returned_param})
        return context.parameters[param_name] if param_name in params else None

    @staticmethod
    def add_params_to_response(
        response: session_pb2.DetectIntentResponse, params: Dict[str, Any], context_name: str,
    ) -> session_pb2.DetectIntentResponse:
        logger_console.info(
            {
                "message": "adding parameter to response",
                "paramter": params,
                "context": context_name,
                "tags": ["parameters", "contexts"],
            }
        )
        parameters = create_parameter_dict(params)
        for j, i in enumerate(response.query_result.output_contexts):
            if i.name == context_name:
                response.query_result.output_contexts[j].parameters.MergeFrom(parameters)
        return response

    @staticmethod
    def delete_param_from_response(response: session_pb2.DetectIntentResponse, param_name: str) -> None:
        logger_console.info(
            {"message": "deleting parameter from response", "parameter": param_name, "tags": ["parameters"]}
        )
        for context in response.query_result.output_contexts:
            if param_name in context.parameters:
                del context.parameters[param_name]


class SingleMessageHandler:
    @staticmethod
    def check_message_for_pattern(message: intent_pb2.Intent.Message, pattern: str) -> bool:
        logger_console.debug({"message": "checking response for text", "content": pattern, "pattern": pattern})
        if message.HasField("text"):
            has_match = SingleMessageHandler._pattern_match_text(message, pattern)
            if has_match:
                return has_match
        if message.HasField("card"):
            has_match = SingleMessageHandler._pattern_match_card(message, pattern)
            if has_match:
                return has_match
        return False

    @staticmethod
    def _pattern_match_text(message: intent_pb2.Intent.Message, pattern: str) -> bool:
        return not not re.findall(pattern, "".join(message.text.text),)

    @staticmethod
    def _pattern_match_card(message: intent_pb2.Intent.Message, pattern: str) -> bool:
        return not not re.findall(pattern, "".join(message.card.subtitle),)

    # TODO: this needs to be improved for more general cases
    @staticmethod
    def get_pattern_from_message(message: intent_pb2.Intent.Message, pattern: str) -> List[str]:
        content: List[str] = []
        if message.HasField("text"):
            content.extend(SingleMessageHandler._get_pattern_from_message_text(message, pattern))
        if message.HasField("card"):
            content.extend(SingleMessageHandler._get_pattern_from_message_card(message, pattern))
        return content

    @staticmethod
    def _get_pattern_from_message_text(message: intent_pb2.Intent.Message, pattern: str) -> List[str]:
        match = re.findall(pattern, message.text.text[0])
        return [i.strip("(").strip(")").strip("'") for i in match]

    @staticmethod
    def _get_pattern_from_message_card(message: intent_pb2.Intent.Message, pattern: str) -> List[str]:
        match = re.findall(pattern, message.card.subtitle)
        return [i.strip("(").strip(")").strip("'") for i in match]

    @staticmethod
    def substitute_pattern_in_message(
        message: intent_pb2.Intent.Message, pattern: str, replace: str, once: bool = False
    ) -> intent_pb2.Intent.Message:
        if message.HasField("text"):
            message = SingleMessageHandler._text_substitution_text(message, pattern, replace, once)
        if message.HasField("card"):
            message = SingleMessageHandler._text_substitution_card(message, pattern, replace, once)
        return message

    @staticmethod
    def _text_substitution_text(
        message: intent_pb2.Intent.Message, pattern: str, replace: str, once: bool
    ) -> intent_pb2.Intent.Message:
        if not len(message.text.text):
            return message
        match: Union[bool, re.Match] = True  # type: ignore
        while match:
            match = re.match(f"(.*)({pattern})(.*)", message.text.text[0])  # type: ignore
            if match:
                new_response = "".join([match.groups()[0], replace, match.groups()[-1]])  # type: ignore
                message.text.text[0] = new_response
                if once:
                    break
        return message

    @staticmethod
    def _text_substitution_card(
        message: intent_pb2.Intent.Message, pattern: str, replace: str, once: bool
    ) -> intent_pb2.Intent.Message:
        match: Union[bool, re.Match] = True  # type: ignore
        while match:
            match = re.match(f"(.*)({pattern})(.*)", message.card.subtitle)  # type: ignore
            if match:
                new_response = "".join([match.groups()[0], replace, match.groups()[-1]])  # type: ignore
                message.card.subtitle = new_response
                if once:
                    break
        return message

    @staticmethod
    def reformat_date_in_message(message: intent_pb2.Intent.Message) -> intent_pb2.Intent.Message:
        if message.HasField("text"):
            message = SingleMessageHandler._reformat_date_text(message)
        if message.HasField("card"):
            message = SingleMessageHandler._reformat_date_card(message)
        return message

    @staticmethod
    def _reformat_date_card(message: intent_pb2.Intent.Message) -> intent_pb2.Intent.Message:
        match: Union[bool, re.Match] = True  # type: ignore
        while match:
            match = re.match(r"(.*)(\d{4}.*T\d\d:\d\d:\d\d)(.*)", message.card.subtitle)  # type: ignore
            if match:
                date = datetime.datetime.fromisoformat(match.groups()[1])  # type: ignore
                new_response = "".join(
                    [match.groups()[0], date.strftime(DATE_FORMAT), match.groups()[2]]  # type: ignore
                )
                message.card.subtitle = new_response
                logger_console.info(f"DATE: {date} formatted to DATE: {date.strftime(DATE_FORMAT)}")
        return message

    @staticmethod
    def _reformat_date_text(message: intent_pb2.Intent.Message) -> intent_pb2.Intent.Message:
        match: Union[bool, re.Match] = True  # type: ignore
        while match:
            match = re.match(r"(.*)(\d{4}.*T\d\d:\d\d:\d\d)(.*)", message.text.text[0])  # type: ignore
            if match:
                date = datetime.datetime.fromisoformat(match.groups()[1])  # type: ignore
                new_response = "".join(
                    [match.groups()[0], date.strftime(DATE_FORMAT), match.groups()[2]]  # type: ignore
                )
                message.text.text[0] = new_response
                logger_console.info(f"DATE: {date} formatted to DATE: {date.strftime(DATE_FORMAT)}")
        return message



import setuptools

with open("README.md", "r") as f:
    long_description = f.read()

setuptools.setup(
    name="ondewo-bpi",
    version="2.0.2",
    author="Ondewo GbmH",
    author_email="info@ondewo.com",
    description="This library starts a proxy for the cai server allowing for fulfillment hooks.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://bitbucket.org/ondewo/ondewo-bpi",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
        "Development Status :: 3 - Alpha",
        "Topic :: Software Development :: Libraries",
    ],
    python_requires=">=3.0.1",
)



# Copyright 2021 ONDEWO GmbH
#
# Licensed under the Apache License, Version 2.0 (the License);
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an AS IS BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict, Optional, List

from grpc._channel import _InactiveRpcError
from ondewo.nlu import context_pb2, session_pb2
from ondewo.nlu.client import Client
from ondewo.logging.logger import logger_console


def add_params_to_cai_context(
    client: Client,
    response: session_pb2.DetectIntentResponse,
    params: Dict[str, Any],
    context: str,
) -> None:
    _add_params_to_cai_context(
        client=client,
        session=get_session_from_response(response),
        params=params,
        context=context,
    )


def _add_params_to_cai_context(
    client: Client,
    session: str,
    params: Dict[str, Any],
    context: str,
) -> None:
    logger_console.info(
        {
            "message": "adding parameter to cai",
            "paramter": params,
            "context": context,
            "tags": ["parameters", "contexts"],
        }
    )
    parameters = create_parameter_dict(params)
    try:
        request = context_pb2.GetContextRequest(name=f"{session}/contexts/{context}")
        existing_context = client.services.contexts.get_context(request)
        existing_context.parameters.MergeFrom(parameters)
        client.services.contexts.update_context(
            request=context_pb2.UpdateContextRequest(
                context=existing_context
            )
        )
    except _InactiveRpcError:
        context = context_pb2.Context(
            name=f"{session}/contexts/{context}",
            lifespan_count=100,
            parameters=parameters,
            lifespan_time=1000
        )
        client.services.contexts.create_context(
            request=context_pb2.CreateContextRequest(
                parent=f'{session}',
                context=context,
            )
        )


def delete_param_from_cai_context(
    client: Client,
    response: session_pb2.DetectIntentResponse,
    param_name: str,
    context: str
) -> None:
    logger_console.info(
        {
            "message": "deleting parameter from cai",
            "paramter": param_name,
            "context": context,
            "tags": ["parameters", "contexts"],
        }
    )
    session = get_session_from_response(response)
    context_name = f"{session}/contexts/{context}"
    existing_context = client.services.contexts.get_context(request=context_pb2.GetContextRequest(name=context_name))
    try:
        del existing_context.parameters[param_name]
        client.services.contexts.delete_context(request=context_pb2.DeleteContextRequest(name=context_name))
        client.services.contexts.create_context(
            request=context_pb2.CreateContextRequest(parent=session, context=existing_context)
        )
    except KeyError:
        logger_console.exception(
            {
                "message": "tried to delete param that didnt exist",
                "parameter": param_name,
                "context": context,
                "tags": ["parameters", "contexts"],
            }
        )


def detect_intent(
    client: Client,
    response: session_pb2.DetectIntentResponse,
    text: str
) -> session_pb2.DetectIntentResponse:
    logger_console.info({"message": "detect intent triggered in bpi helpers", "tags": ["timing"]})
    request = get_detect_intent_request(text=text, session=get_session_from_response(response=response),)
    logger_console.info({"message": "detect intent returned in bpi helpers", "tags": ["timing"]})
    result = client.services.sessions.detect_intent(request)
    logger_console.warning(f"wrote {text}, received {result.query_result.fulfillment_messages}")
    return result


def get_detect_intent_request(
    session: str,
    text: str,
    language: str = 'de-DE',
    query_params: Optional[session_pb2.QueryParameters] = None
) -> session_pb2.DetectIntentRequest:
    request = session_pb2.DetectIntentRequest(
        session=session,
        query_input=session_pb2.QueryInput(text=session_pb2.TextInput(text=text, language_code=language),),
        query_params=query_params,
    )
    return request


def create_parameter_dict(my_dict: Dict) -> Optional[Dict[str, context_pb2.Context.Parameter]]:
    assert isinstance(my_dict, dict) or my_dict is None, "parameter must be a dict or None"
    if my_dict is not None:
        return {
            key: context_pb2.Context.Parameter(
                display_name=key,
                value=my_dict[key]
            )
            for key in my_dict
        }
    return None


# This function creates a detect intent request that will trigger a specific intent
#   using the 'exact intent' trigger
def trigger_intent(
    client: Client,
    session: str,
    intent_name: str,
    language: str = "de-DE",
    additional_contexts: Optional[List[context_pb2.Context]] = None,
) -> session_pb2.DetectIntentResponse:
    """
    Trigger a specific intent in the NLU backend without intent matching.

    Args:
        client: nlu client
        session: full session to perform the trigger in ('parent/<PROJECT_ID>/agent/sessions/<SESSION_ID>')
        intent_name: intent that you want to trigger
        language: language of the project
        additional_contexts: if you want to add additional contexts to the session

    Returns:
        session_pb2.DetectIntentResponse
    """
    if not additional_contexts:
        additional_contexts = []

    logger_console.warning({"message": "triggering specific intent", "intent_name": intent_name})
    trigger_context = create_context_struct(
        context=f"{session}/contexts/exact_intent",
        parameters=create_parameter_dict({"intent_name": intent_name}),
        lifespan_count=1,
    )
    request = get_detect_intent_request(
        text=f"Triggering Specific Intent: {intent_name}",
        session=session,
        language=language,
        query_params=session_pb2.QueryParameters(contexts=[trigger_context, *additional_contexts]),
    )
    result = client.services.sessions.detect_intent(request)
    logger_console.warning(f"triggered {intent_name}")
    return result


def create_context_struct(context: str, parameters: Optional[Dict[str, context_pb2.Context.Parameter]], lifespan_count: int = 5) -> context_pb2.Context:
    context_struct: context_pb2.Context = context_pb2.Context(
        name=f"{context}", lifespan_count=lifespan_count, parameters=parameters
    )
    return context_struct


# This function deletes periods from the text in a request
def strip_final_periods_from_request(request: session_pb2.DetectIntentRequest,) -> session_pb2.DetectIntentRequest:
    stripped = request.query_input.text.text.strip(".")
    request.query_input.text.text = stripped
    return request


def get_session_from_response(response: session_pb2.DetectIntentResponse) -> str:
    return response.query_result.diagnostic_info["sessionId"]   # type: ignore



# Copyright 2021 ONDEWO GmbH
#
# Licensed under the Apache License, Version 2.0 (the License);
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an AS IS BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os
from typing import Dict, Optional, Tuple

from dotenv import load_dotenv
from ondewo.nlu.client import Client
from ondewo.nlu.client_config import ClientConfig
from ondewo.logging.logger import logger, logger_console

import ondewo_bpi.__init__ as file_anchor

parent = os.path.abspath(os.path.join(os.path.dirname(file_anchor.__file__), os.path.pardir))


load_dotenv("./bpi.env")

PORT: str = os.getenv("PORT", "50051")
CAI_HOST: Optional[str] = os.getenv("CAI_HOST")
CAI_PORT: Optional[str] = os.getenv("CAI_PORT")
CAI_TOKEN: Optional[str] = os.getenv("CAI_TOKEN")
HTTP_AUTH_TOKEN: Optional[str] = os.getenv("HTTP_BASIC_AUTH")
USER_NAME: Optional[str] = os.getenv("USER_NAME")
USER_PASS: Optional[str] = os.getenv("USER_PASS")
SECURE: Optional[str] = os.getenv("SECURE")

config_path: str = os.getenv("CONFIG_PATH", "/home/ondewo/config.json")

client_configuration_str = (
    "\nnlu-client configuration:\n"
    + f"   Secure: {SECURE}\n"
    + f"   Host: {CAI_HOST}\n"
    + f"   Port: {CAI_PORT}\n"
    + f"   Http_token: {HTTP_AUTH_TOKEN}\n"
    + f"   User_name: {USER_NAME}\n"
    + f"   Password: {USER_PASS}\n"
)
logger_console.info(client_configuration_str)


class CentralClientProvider:
    """
    provide a central nlu-client instance to the bpi server without building it on import
    """

    def __init__(self) -> None:
        self.config = None
        self.client = None
        self._built = False

    def instantiate_client(self, cai_port: str = "") -> Tuple[ClientConfig, Client]:
        if cai_port == "":
            trial_port = CAI_PORT
            if trial_port == "" or not trial_port:
                trial_port = "50055"
            cai_port = trial_port

        if SECURE == "True":
            with open(config_path, "r") as fi:
                json_dict: Dict = json.load(fi)

            logger.info("configuring secure connection")
            config: ClientConfig = ClientConfig(
                host=CAI_HOST,
                port=cai_port,
                http_token=HTTP_AUTH_TOKEN,
                user_name=USER_NAME,
                password=USER_PASS,
                grpc_cert=json_dict["grpc_cert"],
            )
            client = Client(config=config)
        else:
            logger.info("configuring INSECURE connection")
            config = ClientConfig(
                host=CAI_HOST, port=cai_port, http_token=HTTP_AUTH_TOKEN, user_name=USER_NAME, password=USER_PASS,
            )
            client = Client(config=config, use_secure_channel=False)
        return config, client

    def get_client(self, cai_port: str = "") -> Client:
        if not self._built:
            self.config, self.client = self.instantiate_client(cai_port=cai_port)
            self._built = True
        return self.client
